open Term
open Result
open Monad.MonadList

(*
module M = Monad.ListMonad

let flatten ac rs =
  M.let* s,t = rs in
  [Ac_subst.flatten ac s, Ac_subst.flatten ac t]
*)

let flatten ac rs =
  List.map
    (fun (s,t) ->
      Ac_subst.flatten ac s, Ac_subst.flatten ac t)
   rs

let inverse  es =
  List.map
    (fun (l,r) -> (r,l))
    es
let inverse' es =
  List.map
    (fun (l,s,r) -> r,s,l)
    es



(** C-modulo **)

(* heuristics for C-terminating TRS *)
let emb_multi c rs =
  let mul = Rewriting.multistep c in
  let* (l1,r) = rs in
  let* l2 = mul l1 in
  [l2,r]

let c_eqaul_by rs p s t =
  not (List.exists (p s) (Rewriting.multistep rs t))

let is_C c (l,r) =
  List.mem (root l) c && Acrpo.rule_C (l,r)

(* divide TRSs to common C-rules and others *)
let common_c r s =
  (* extract symbols instead of rules;
     because rules of R and S may not share variables *)
  let r_c_sym  =Listset.diff (Acrpo.c_sym r) (Acrpo.ac_sym r) in
  let s_c_sym  =Listset.diff (Acrpo.c_sym s) (Acrpo.ac_sym s) in
  let c_sym = Listset.inter r_c_sym s_c_sym in
  if c_sym = [] then
    None
  else
    let c = List.filter (is_C c_sym) r in
    Some ((Listset.diff r c, Listset.diff s c), c_sym, c)

(** AC-modulo **)

let common_ac r s =
  match Ac_rewriting.find_ac r, Ac_rewriting.find_ac s with
  | Some (rs1, ac_sym , ac1, ac_all),
    Some (rs2, ac_sym', ac2, _) when ac_sym = ac_sym'
      -> Some ((rs1,rs2), ac_sym, ac_all)
  | _ -> None

let common_acS r s = if !Ref.jkac then common_ac r s else None

let common_cS r s = if !Ref.jkc then common_c r s else None

(* check commutation modulo of AC cup C systems s.t. AC cap C = [] *)
let ac_jk86_modulo r s =
  let (rr,ss), c_sym, c_all = 
    match common_cS r s with
    | None   -> (r,s),[],[]
    | Some x -> x in
  (* failure if TRSs has different AC rules or non terminating *)
  let ((rr1,ss1), ac_sym, ac_all) =
    match common_acS rr ss with
    | None -> (rr,ss),[],[]
    | Some x -> x in
  (* check E-termination (We have only AC-terminating module) *)
  if not (Acrpo.terminate (c_sym@ac_sym) (rr1@ss1)) then
    MAYBE
  else
    let rc,sc = emb_multi c_all rr1, emb_multi c_all ss1 in
    (* peak   <-R- . -S-> *)
    let cps =
      Listx.unique (
          Ac_overlap.critical_peak2 ac_sym rc (sc@ac_all)
          @ inverse' (Ac_overlap.critical_peak2 ac_sym sc (rc@ac_all))) in
    (* valley -S-> . <-R- *)
    let cps_nf =
      List.map
        (fun (s,u,t) ->
          Ac_rewriting.nf ac_sym sc s, u, Ac_rewriting.nf ac_sym rc t)
        cps in
    let p s t = Ac_rewriting.equal ac_sym s t in
    let nj =
      List.filter
        (fun (s,u,t) -> c_eqaul_by c_all p s t)
        cps_nf in
    match nj with
    | []             -> YES ()
    | (s, u, t) :: _ -> NO (s,u,t)


(* convert results from the implication CRM_AC => CR *)
let ac_jk86 r s =
  match ac_jk86_modulo r s with
    YES x -> YES x
  | NO _  -> MAYBE
  | MAYBE -> MAYBE


(** A-modulo **)

let is_A a (l,r) =
  List.mem (root l) a && Acrpo.rule_A (l,r)

(* common C-symbols and rules without A *)
let common_a r s =
  (* systems have same A-rules without AC-rules? *)
  let a_sym, a_sym' = Acrpo.a_sym r, Acrpo.a_sym s in
  if Listset.union a_sym a_sym' = [] || a_sym <> a_sym' ||
     Listset.inter a_sym (Acrpo.ac_sym (r@s)) <> [] then
    None
  else
    (* all A-rules are reversible? *)
    let a_all = Listx.concat_map Acrpo.a_class a_sym in
    let not_A rs = List.filter (fun lr -> not (is_A a_sym lr)) rs in
    if List.for_all
      (fun rule -> List.exists (Match.variant_rule rule) (Listset.inter r s))
      a_all
    then
      Some
        ((not_A r, not_A s) , a_sym , a_all)
    else
      None

let common_aS r s = if !Ref.jka then common_a r s else None

let a_unifiers a_sym ((l1,r1),p,(l2,r2)) =
  let var (x,y) = Term.variables x @ Term.variables y in
  let at = subterm_at p l2 in
  if Listset.inter (var (l1,r1)) (var (l2,r2)) = [] && not (is_variable at)
     (* exclusion of variant rules is unsound for extended critical pairs *)
     (* && (p <> [] || not (Match.variant_rule (l1,r1) (l2,r2))) *)
  then
    Assoc.unify a_sym l1 at
  else
    []

let a_overlaps a_sym r s0 =
  let s =
    if Listset.inter (Trs.variables r) (Trs.variables s0) = [] then s0 else Trs.rename s0
  in
  let* lr1 = r in
  let* l2,r2 as lr2 = s in
  let* p = function_positions l2 in
  let* m = a_unifiers a_sym (lr1,p,lr2) in
  [(lr1,p,lr2), m]

let a_cpp2 a_sym r s =
  List.map
    (fun (((l1,r1),p,(l2,r2)), m) ->
      replace (Term.substitute m l2) (Term.substitute m r1) p,
      Term.substitute m l2,
      Term.substitute m r2)
    (a_overlaps a_sym r s)

let rec rules_with_defined_sym = function
    [] -> []
  | ((F (f, _) as l), r) :: rs -> (f,l,r) :: rules_with_defined_sym rs
  | _ -> invalid_arg "Ac_cr.rules_with_defined_sym"


(* ExtA(R) = 
     { f(l,x) -> f(r,x), 
       f(x,l) -> f(x,r),
       f(x,f(l,y)) -> f(x,f(r,y)) 
     | l -> r in R with root(l) = f in F_A}
*)
let a_ext a_sym rs = 
  let x = Term.new_var () in
  let y = Term.new_var () in
  let exts =
    let* (f, l, r) = rules_with_defined_sym rs in
    guard (List.mem f a_sym) >>= fun () ->
    [F (f, [x; l]), F (f, [x; r]);
     F (f, [l; x]), F (f, [r; x]);
     F (f, [x; F (f, [l; y])]), F (f, [x; F (f, [r; y])]) ]
  in
  rs @ exts


let a_c_jk86_modulo r s =
  (* failure if TRSs has different AC rules or non terminating *)
  let ((rr0,ss0),c_sym,c_all) =
    match common_cS r s with
    | None   -> (r,s),[],[]
    | Some x -> x in
  let ((rr,ss), a_sym, a_all) =
    match common_aS rr0 ss0 with
    | None   -> (rr0,ss0),[],[]
    | Some x -> x in
  (* check E-termination (We have only AC-terminating module) *)
    if not (Acrpo.terminate (c_sym@a_sym) (rr@ss)) then
        MAYBE
    else
      let rc,sc = emb_multi c_all rr, emb_multi c_all ss in
      let ext_rc = a_ext a_sym rc in
      let ext_sc = a_ext a_sym sc in
      (* peak   <-R- . -S-> *)
      let cpps =
          (a_cpp2 a_sym ext_rc (ext_sc@a_all) @
           inverse' (a_cpp2 a_sym ext_sc (ext_rc@a_all))) in
      (* valley -S-> . <-R- *)
      let cpps_nf =
        List.map
          (fun (t,s,u) ->
            Assoc.nf a_sym ext_sc t, s, Assoc.nf a_sym ext_rc u)
          cpps in
      let p s t = Assoc.equal a_sym s t in
      let nj =
        let* (t,s,u) = cpps_nf in
        let* (t',u') = Trs.rename [t,u] in
        guard (c_eqaul_by c_all p t' u') >>= fun () ->
        [ t,s,u ] in
      match nj with
      | []           -> YES ()
      | (t,s,u) :: _ -> NO (t,s,u)

(* convert results from the implication CRM_A <=> CR *)
let a_c_jk86 r s =
  a_c_jk86_modulo r s


let jk86 rs ss =
  Result.exists
    (* (fun thm -> thm (Step.reduce rs) (Step.reduce ss)) <- integrate to main*)
    (fun thm -> thm rs ss)
    (* [ ac_jk86; a_jk86 ] *)
    [ a_c_jk86; ac_jk86 ]
